#!/usr/bin/python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
#
# Extract ICONs from MZ/NE MS Windows executable file
#

from struct import *

with open('../bin/data/DUNE2.ICO', 'rb') as content_file:
    content = content_file.read()

icon_header = {}
icon_data = {}

# RT_GROUP_ICON = 14 = 0x0e
# RT_ICON = 3
if content[0:2] == 'MZ':
    newoffset, = unpack_from('<H', content, 0x3C)
    if content[newoffset:newoffset+2] == 'NE':
        resoffset, = unpack_from('<H', content, newoffset+0x24)
        ressegcount, = unpack_from('<H', content, newoffset+0x34)
        #print resoffset, ressegcount
        rscAlignShift, = unpack_from('<H', content, newoffset+resoffset)
        offset = newoffset+resoffset+2
        while True:
            rtTypeID, rtResourceCount, rtReserved = unpack_from('<HHI', content, offset)
            if rtTypeID == 0:
                break
            offset = offset + 8
            print 'resource type 0x%04x count %d :' % (rtTypeID, rtResourceCount)
            for j in range(0, rtResourceCount):
                rnOffset, rnLength, rnFlags, rnID, rnHandle, rnUsage = unpack_from('<HHHHHH', content, offset)
                offset = offset + 12
                realoffset = rnOffset << rscAlignShift
                reallen = rnLength << rscAlignShift
                print '  offset=0x%06x (0x%04x) len=%d (0x%04x) flags=%04x ID=%04x' % (realoffset, rnOffset, reallen, rnLength, rnFlags, rnID)
                if rtTypeID == 0x800e:
                    icon_header[rnID] = content[realoffset:realoffset+18] + '\x16\x00\x00\x00'
                elif rtTypeID == 0x8003:
                    icon_data[rnID] = content[realoffset:realoffset+reallen]

for icon_id in icon_header.keys():
    filename = '%04x.ICO' % icon_id
    with open(filename, 'wb') as icon_file:
        icon_file.write(icon_header[icon_id])
        icon_file.write(icon_data[icon_id])
        print '%s written' % filename




#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab

def decode_format80(data):
	dest = ''
	i = 0
	while i < len(data) and ord(data[i]) != 0x80:
		cmd = ord(data[i])
		i += 1
		if cmd == 254:	# 1111 1110 SS SS : long set
			size = ord(data[i]) + (ord(data[i+1]) << 8)
			#print cmd, 'long set', size, data[i+2].encode('hex')
			dest += (data[i+2] * size)
			i += 3
		elif cmd == 255:	# 1111 1111 SS SS OO OO : long absolute move
			size = ord(data[i]) + (ord(data[i+1]) << 8)
			offset = ord(data[i+2]) + (ord(data[i+3]) << 8)
			#print cmd, 'long absolute move', offset, size
			while size > 0:
				dest += dest[offset]
				offset += 1
				size -= 1
			i += 4
		elif (cmd & 128) == 0:	# 0sss oooo OO : short relative move
			size = (cmd >> 4) + 3
			offset = ((cmd & 15) << 8) + ord(data[i])
			i += 1
			#print cmd, 'short relative move', offset, size
			while size > 0:
				dest += dest[-offset]
				size -= 1
		elif (cmd & 64) != 0:	# 11ss ssss OO OO : short absolute move
			size = (cmd & 63) + 3
			offset = ord(data[i]) + (ord(data[i+1]) << 8)
			#print cmd, 'short absolute move', offset, size
			while size > 0:
				dest += dest[offset]
				offset += 1
				size -= 1
			i += 2
		else:	# 10ss ssss : short copy
			size = cmd & 63
			#print cmd, 'short copy', size
			dest += data[i:i+size]
			i += size
	return dest

def decode_format40(data):
	dest = ''
	i = 0
	while i < len(data):
		cmd = ord(data[i])
		i += 1
		if cmd == 0:
			# RLE
			dest += (data[i+1] * ord(data[i]))
			i += 2
		elif (cmd & 0x80) == 0:
			# copy
			dest += data[i:i+cmd]
			i += cmd
		elif cmd == 0x80:
			# 16 bit command
			cmd = ord(data[i]) + (ord(data[i+1]) << 8)
			i += 2
			if cmd == 0:
				# end
				break
			elif (cmd & 0x8000) == 0:
				# skip
				dest += (chr(0) * cmd)
			elif (cmd & 0x4000) == 0:
				count = cmd & 0x3fff
				dest += data[i:i+count]
				i += count
			else:
				count = cmd & 0x3fff
				dest += (data[i] * count)
				i += 1
		else:
			# skip
			dest += (chr(0) * (cmd & 0x7f))
	return dest




#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
# read WSA (WestWood Studios Animations) files

from struct import *
import sys
from westwood_codecs import *
from save_pictures import *

def load_palette(filename):
	with open(filename, 'rb') as pal_file:
		return ''.join(map(lambda c: chr((ord(c) << 2) + (ord(c) >> 6)), list(pal_file.read())))

def xor_strings(s1, s2):
	dest = ''
	for i in range(min(len(s1),len(s2))):
		dest += chr(ord(s1[i]) ^ ord(s2[i]))
	return dest

# frame : set to last frame of previous WSA for "continuation" WSA
def extract_wsa(filename, palette = None, frame = None):
	with open(filename, 'rb') as wsa_file:
		wsa = wsa_file.read()
		(frames, width, height, req_buffer_size, has_palette) = unpack_from('<HHHHH', wsa)
		print '%s length %d (0x%06X)' % (filename, len(wsa), len(wsa))
		print '  %2d frames %3dx%3d req_buffer_size=%5d has_palette=%d' % (frames, width, height, req_buffer_size, has_palette)
		offset = 10
		offsets = []
		for i in range(frames+2):
			(offset_frame,) = unpack_from('<L', wsa, offset)
			offsets += [offset_frame]
			offset += 4
		print ' ', ':'.join(map(lambda i : '%06X' % i, offsets))
		datablocks = [ decode_format80(wsa[offsets[i]:offsets[i+1]]) if offsets[i] != 0 else '' for i in range(frames+1) ]
		print ' ', map(lambda e : len(e), datablocks)
		pictures = map(decode_format40, datablocks)
		print ' ', map(lambda e : len(e), pictures)
		frames = []
		if len(pictures[0]) > 0:
			pixels = pictures[0]
			frames.append(pixels)
		else:
			if frame is None:
				pixels = chr(0) * (width * height)
				print 'no first frame provided'
			else:
				# continue from last frame of previous WSA
				pixels = frame
				if len(pixels) != (width * height):
					print 'frame size mismatch : ', len(pixels), width * height
		for i in range(1, len(pictures)):
			pbm_filename = '%s_%02d.pbm' % (filename, i)
			if len(pictures[i]) > 0:
				pixels = xor_strings(pixels, pictures[i])
				#save_pbm(pbm_filename, width, height, pixels, palette)
				frames.append(pixels)
		save_flc(filename + '.flc', width, height, frames, palette)
		return frames[-1]

if len(sys.argv) <= 1:
	print "usage : %s [-p palette.PAL] file.WSA" % sys.argv[0]
else:
	args = sys.argv[1:]
	frame = None
	palette = None
	if args[0] == '-p':
		palette = load_palette(args[1])
		args = args[2:]
	for filename in args:
		frame = extract_wsa(filename, palette, frame)



#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
# read CPS files

from struct import *
import sys
from westwood_codecs import *
from save_pictures import *

palette = None

def load_palette(filename):
	global palette
	with open(filename, 'rb') as pal_file:
		palette = ''.join(map(lambda c: chr((ord(c) << 2) + (ord(c) >> 6)), list(pal_file.read())))

def extract_cps(filename):
	global palette
	with open(filename, 'rb') as cps_file:
		cps = cps_file.read()
		(packed_size, encoding, decoded_size, offset) = unpack_from('<HHLH', cps)
		print filename, packed_size, encoding, decoded_size, offset
		if offset > 0:
			pal = ''.join(map(lambda c: chr((ord(c) << 2) + (ord(c) >> 6)), list(cps[10:10+offset])))
		else:
			pal = palette
		if encoding == 0:
			pixels = cps[10+offset:2+packed_size]
		elif encoding == 4:	# format80
			pixels = decode_format80(cps[10+offset:2+packed_size])
		else:
			print 'Unknown encoding' 
			return
		save_pbm(filename + '.pbm', 320, 200, pixels, pal)

if len(sys.argv) <= 1:
	print "usage : %s [-p palette.PAL] file.CPS" % sys.argv[0]
else:
	args = sys.argv[1:]
	if args[0] == '-p':
		load_palette(args[1])
		args = args[2:]
	for filename in args:
		extract_cps(filename)



#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
# read FNT font files from Dune II

from struct import *
import sys
from save_pictures import *

def extract_fnt(filename):
	with open(filename, 'rb') as fnt_file:
		fnt = fnt_file.read()
		(size, sig, params, offsets, width_list, char_data, line_list) = unpack_from('<HHHHHHH', fnt)
		print filename, size, sig, params, offsets, width_list, char_data, line_list
		(p1, p2, p3, max_index, height, max_width) = unpack_from('<BBBBBB', fnt, params)
		#print fnt[params:offsets].encode('hex')
		print ' ', p1, p2, p3, max_index, height, max_width
		lines = [''] * height
		for i in range(max_index + 1):
			width = ord(fnt[width_list+i])
			(unused_lines, used_lines) = unpack_from('<BB', fnt, line_list + i * 2)
			(offset,) = unpack_from('<H', fnt, offsets + i * 2)
			#print i, width, unused_lines, used_lines, offset
			for y in range(height):
				if y < unused_lines or y >= unused_lines + used_lines:
					line = chr(0) * width
				else:
					line = ''
					for x in range(width):
						data = ord(fnt[offset])
						if (x & 1) != 0:
							data = data >> 4
						else:
							data = data & 0x0f
						line += chr(data)
						if (x & 1) != 0 or x == width - 1:
							offset += 1
				lines[y] += line
		save_pbm(filename + '.pbm', len(lines[0]), height, ''.join(lines), None)

if len(sys.argv) <= 1:
	print "usage : %s file.FNT" % sys.argv[0]
else:
	args = sys.argv[1:]
	for filename in args:
		extract_fnt(filename)



#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
# extract sprites from SHP files (Westwood studios games)

from struct import *
import sys
from westwood_codecs import *
from save_pictures import *

palette = None

def load_palette(filename):
	global palette
	with open(filename, 'rb') as pal_file:
		palette = ''.join(map(lambda c: chr((ord(c) << 2) + (ord(c) >> 6)), list(pal_file.read())))

def decode_sprite_data(data, palette):
	# RLE 00 nn = nn transparent pixels (00)
	dest = ''
	i = 0
	while i < len(data):
		if ord(data[i]) == 0:
			dest += data[i] * ord(data[i+1])
			i += 2
		else:
			dest += data[i] if palette is None else palette[ord(data[i])]
			i += 1
	return dest
	
def decode_sprite(data):
	palette = None
	(flags, height, width, height2, datalen, decodedlen) = unpack_from('<HBHBHH', data)
	print '0x%04x %02dx%02d %02d %d %d' % (flags, width, height, height2, datalen, decodedlen)
	offset = 10
	if (flags & 0x1) != 0:	# local 16 color palette
		palette = data[offset:offset+16]
		offset += 16
	if (flags & 0x2) == 0:
		sprite_data = decode_format80(data[offset:])
	else:
		sprite_data = data[offset:]
	pixels = decode_sprite_data(sprite_data, palette)
	print 'decodedlen =', len(sprite_data), 'pixels =', len(pixels)
	for line in map(lambda s : s.encode('hex'), split_string(pixels, width)):
		print line
	return (width, height, pixels)

def extract_shp(filename):
	print "Extracting sprites from", filename, ':'
	with open(filename, 'rb') as shp_file:
		shp = shp_file.read()
		(count,) = unpack_from('<H', shp, 0)
		print ' ', len(shp), 'bytes, ', count, "icons :"
		offsets = []
		for i in range(count+1):
			(offset,) = unpack_from('<L', shp, 2 + i*4)
			offsets.append(offset+2)
		if offsets[0] != 6 + count * 4 or offsets[count] != len(shp):
			#print 'error', offsets
			offsets = []	# another format (2 bytes per offset)
			for i in range(count+1):
				(offset,) = unpack_from('<H', shp, 2 + i*2)
				offsets.append(offset)
		for i in range(count):
			#print '   ', i, offsets[i]
			save_pbm(*("%s_%03d.pbm" % (filename, i),) + decode_sprite(shp[offsets[i]:offsets[i+1]]) + (palette,))

if len(sys.argv) <= 1:
	print "usage : %s [-p palette.PAL] file.shp" % sys.argv[0]
else:
	args = sys.argv[1:]
	if args[0] == '-p':
		load_palette(args[1])
		args = args[2:]
	for filename in args:
		extract_shp(filename)



#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#

from struct import pack

def split_string(s, n):
	return [s[i*n:i*n+n] for i, j in enumerate(s[::n])]

def save_pbm(filename, width, height, pixels, palette):
	if (width & 1) != 0:
		pixels = ''.join([line + chr(0) for line in split_string(pixels, width)])
	with open(filename, 'wb') as pbm_file:
		bmhd_chunk = 'BMHD' + pack('>LHHHHBBBBHBBHH', 20, width, height, 0, 0, 8, 2, 0, 0, 0, 10, 10, width, height);
		cmap_chunk = '' if not palette else 'CMAP' + pack('>L', len(palette)) + palette
		body_chunk = 'BODY' + pack('>L', len(pixels)) + pixels

		size = 4 + len(body_chunk) + len(bmhd_chunk) + len(cmap_chunk)
		form = 'FORM' + pack('>L', size) + 'PBM ' + bmhd_chunk + cmap_chunk + body_chunk
		pbm_file.write(form)
		print filename, "written"

def save_flc(filename, width, height, frames, palette):
	with open(filename, 'wb') as flc_file:
		frame_delay = 100	# ms per frame
		# 1st frame
		flc_frames = []
		if palette is None:
			palette_sub_chunk = None
		else:
			# sub_chunk_size, sub_chunk_type
			# packet_count, skip, count (count = 0 for 256)
			palette_sub_chunk = pack('<LHHBB', len(palette) + 6 + 2 + 2, 0x04, 1, 0, (len(palette) / 3) & 0xff)
			palette_sub_chunk += palette
		# Full frame RLE : 0x0f
		encoded_frame = ''
		for y in range(height):
			packets = []
			x = 1
			count = 1
			repeat = True
			#print frames[0][y*width:(y+1)*width].encode('hex')
			while x < width:
				#print x, y, repeat
				if repeat:
					if frames[0][y*width+(x-1)] == frames[0][y*width+x] and count < 127:
						count += 1
					elif count > 2:
						packets.append(chr(count)+frames[0][y*width+(x-1)])
						count = 1
					else:
						repeat = False
						count += 1
				else:
					if frames[0][y*width+(x-2)] == frames[0][y*width+(x-1)] == frames[0][y*width+x]:
						packets.append(pack('<b', 2-count)+frames[0][y*width+x-count:y*width+x-2])
						repeat = True
						count = 3
					elif count >= 128:	# To be checked
						packets.append(pack('<b', 1-count)+frames[0][y*width+x-count:y*width+x-1])
						count = 2
					else:
						count += 1
				x += 1
			if count > 0:
				if repeat:
					packets.append(chr(count)+frames[0][y*width+(x-1)])
				else:
					packets.append(pack('<b', -count)+frames[0][y*width+x-count:y*width+x])
			#print count, map(lambda s : s.encode('hex'), packets)
			encoded_frame += pack('<B', len(packets)) + ''.join(packets)
		full_frame_rle_sub_chunk = pack('<LH', len(encoded_frame) + 6, 0x0f) + encoded_frame
		if palette_sub_chunk is None:
			first_frame = pack('<LHHHHHH', len(full_frame_rle_sub_chunk) + 16, 0xf1fa, 1, 0, 0, 0, 0)
		else:
			first_frame = pack('<LHHHHHH', len(full_frame_rle_sub_chunk) + len(palette_sub_chunk) + 16, 0xf1fa, 2, 0, 0, 0, 0)
			first_frame += palette_sub_chunk
		first_frame += full_frame_rle_sub_chunk
		flc_frames = [first_frame]
		for frame in range(1, len(frames)):
			encoded_frame = ''
			equals_lines = 0
			line_count = 0
			for y in range(height):
				current_line = frames[frame][y*width:(y+1)*width]
				previous_line = frames[frame-1][y*width:(y+1)*width]
				if current_line == previous_line:
					equals_lines += 1
				else:
					if (equals_lines > 0):
						encoded_frame += pack('<h', -equals_lines)
						equals_lines = 0
					packets = []
					# no compression test:
					#if width > 254:
					#	packets.append(pack('<Bb', 0, 127) + frames[frame][y*width:y*width+254])
					#	packets.append(pack('<Bb', 0, (width/2)-127) + frames[frame][y*width+254:(y+1)*width])
					#else:
					#	packets.append(pack('<Bb', 0, (width/2)) + frames[frame][y*width:(y+1)*width])
					x = 0
					while x < width:
						skip = 0
						count = 0
						while x < width and skip < 254 and current_line[x:x+2] == previous_line[x:x+2]:
							x += 2
							skip += 2
						rle = True
						while x < width and count < 127 and current_line[x:x+2] != previous_line[x:x+2]:
							if rle and count > 0 and current_line[x:x+2] != current_line[x-2:x]:
								rle = False
							x += 2
							count += 1
						if rle and count > 1:
							packets.append(pack('<Bb', skip, -count) + current_line[x-2:x])
						else:
							packets.append(pack('<Bb', skip, count) + current_line[x-count*2:x])
					encoded_frame += pack('<H', len(packets)) + ''.join(packets)
					line_count += 1
			# frame with 1 sub chunk of type 7 (DELTA_FLC)
			flc_frames.append(pack('<LHHHHHHLHH', len(encoded_frame) + 24, 0xf1fa, 1, 0, 0, 0, 0, len(encoded_frame) + 8, 0x07, line_count) + encoded_frame)
		flc_data = ''.join(flc_frames)
		# 0xAF12 is signature for FLC files
		flc_header = pack('<LHHHHHHL', len(flc_data) + 128, 0xAF12, len(flc_frames), width, height, 8, 0, frame_delay)
		flc_header += chr(0) * (128 - len(flc_header))
		flc_file.write(flc_header)
		flc_file.write(flc_data)
		print filename, "written"



#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
# read ICON.ICN file (IFF)

from struct import *
import sys
from westwood_codecs import *
from save_pictures import *

def load_palette(filename):
	with open(filename, 'rb') as pal_file:
		return ''.join(map(lambda c: chr((ord(c) << 2) + (ord(c) >> 6)), list(pal_file.read())))

def load_icon_map(filename):
	with open(filename, 'rb') as map_file:
		s = map_file.read()
		return [unpack_from('<H', s, i)[0] for i in range(0, len(s), 2)]

def parse_iff(data):
	form = data[0:4]
	if form != 'FORM':
		return None
	(form_len,) = unpack_from('>L', data, 4)
	if ((form_len + 9)&~1) != ((len(data) + 1)&~1):
		print "FORM len mismatch", form_len + 8, len(data)
	iff_type = data[8:12]
	print iff_type, form_len
	offset = 12
	iff_data = {}
	while offset <= len(data)-8:
		chunk = data[offset:offset+4]
		(chunk_size,) = unpack_from('>L', data, offset + 4)
		print '%06X %s %6d' % (offset, chunk, chunk_size)
		offset += 8
		iff_data[chunk] = data[offset:offset+chunk_size]
		offset += (chunk_size + 1) & ~1
	return (iff_type, iff_data)

def decode_sset(data):
	(encoding,size,offset) = unpack_from('<HLH', data)
	offset += 8
	if encoding == 0:
		return data[offset:offset+size]
	elif encoding == 4:	# format80
		return decode_format80(data[offset:])

def decode_tile(data, pal, width, height):
	rows = []
	for y in range(height):
		row = ''
		for x in range(width/2):
			v = ord(data[y*(width/2)+x])
			left = pal[v >> 4]
			right = pal[v & 0x0f]
			row += left
			row += right
		#print row.encode('hex')
		rows.append(row)
	return rows

def save_tileset(filename, tiles, width, height, palette, tileset_width = 320):
	count = len(tiles)
	tileset_height = height * ((count + (tileset_width / width) - 1) / (tileset_width / width))
	#print tileset_width, tileset_height
	step = tileset_width / width
	pixels = ''
	for i in range(0, count, step):
		chunk = reduce(lambda a, b: map(lambda x, y: x+y, a, b), tiles[i:i+step])
		chunk = map(lambda s: s if len(s) >= tileset_width else s + (chr(0)*(tileset_width-len(s))), chunk)
		pixels += ''.join(chunk)
	save_pbm(filename, tileset_width, tileset_height, pixels, palette)

def extract_icn(filename, palette = None, icon_map = None):
	with open(filename, 'rb') as icn_file:
		icn = icn_file.read()
		(icn_type, icn_data) = parse_iff(icn)
		width = ord(icn_data['SINF'][0]) * 8
		height = ord(icn_data['SINF'][1]) * 8
		tile_bytes = width * height / 2
		pixel_data = decode_sset(icn_data['SSET'])
		count = len(pixel_data) / tile_bytes
		print '%d tiles of %dx%d pixels, %d bytes' % (count, width, height, tile_bytes)
		if len(icn_data['RTBL']) != count:
			print 'RTBL size mismatch :', count, len(icn_data['RTBL'])
		else:
			tiles = []
			for i in range(count):
				pal_index = ord(icn_data['RTBL'][i])
				pal = icn_data['RPAL'][pal_index*16:(pal_index+1)*16]
				tile = decode_tile(pixel_data[tile_bytes*i:tile_bytes*(i+1)], pal, width, height)
				istransparent = False
				for line in tile:
					for j in range(len(line)):
						if ord(line[j]) == 0:
							istransparent = True
							break
				if istransparent:
					print 'Tile %3d (palette %2d) is transparent %02x' % (i, pal_index, ord(pal[0]))
				tiles.append(tile)
			save_tileset(filename + '.pbm', tiles, width, height, palette)
			if icon_map is not None:
				count = icon_map[0]
				for i in range(count):
					if icon_map[i+1] == 0:
						icons = icon_map[icon_map[i]:]
					else:
						icons = icon_map[icon_map[i]:icon_map[i+1]]
					#print i, icon_map[i], len(icons), icons
					if len(icons) > 60 and i != 20:
						tileset_width = 10 * width
					elif (len(icons) % 3) == 0 and i != 12 and i != 26:
						tileset_width = 3 * width
					elif (len(icons) % 2) == 0:
						tileset_width = 2 * width
					elif len(icons) < 4:
						tileset_width = width * len(icons)
					else:
						tileset_width = 4 * width
					save_tileset("%s_%02d.pbm" % (filename, i), [tiles[j] for j in icons], width, height, palette, tileset_width)
			palette_usage = ['0'] * 256
			for c in icn_data['RPAL']:
				palette_usage[ord(c)] = '1'
			#print ''.join(palette_usage)

if len(sys.argv) <= 1:
	print "usage : %s [-p palette.PAL] [-m icon.MAP] file.ICN" % sys.argv[0]
else:
	palette = None
	icon_map = None
	args = sys.argv[1:]
	while len(args) > 2 and args[0][0] == '-':
		if args[0] == '-p':
			palette = load_palette(args[1])
			args = args[2:]
		elif args[0] == '-m':
			icon_map = load_icon_map(args[1])
			args = args[2:]
	for filename in args:
		extract_icn(filename, palette, icon_map)
